@@ -1,6 +1,7 @@
 import os
 import sys
 import json
+import time
 import shlex
 import asyncio
 
@@ -33,7 +34,7 @@ parser.add_argument('--timeout', help='llama.cpp timeout in seconds', default=30
 parser.add_argument('--backend', help='llama.cpp execution backend', default='cpu', type=str, choices=['cpu', 'clblast', 'hipblas', 'cublas'])
 parser.add_argument('--models-path', help='models directory path', default='models')
 parser.add_argument('--llama-cpp-path', help='llama.cpp directory path', default='llama.cpp')
-parser.add_argument('--platforms-devices', help='Custom platforms and devices indexes, example: 0:0,0:1', type=str, required=False)
+parser.add_argument('--platforms-devices', help='Custom platforms and devices indexes, example: 0:0,0:1', default="0:0")
 cli_args = parser.parse_args()
 
 HOST = cli_args.host
@@ -83,6 +84,33 @@ def init_devices():
     print('devices_locks:', devices_locks)
     # print('devices_procs:', devices_procs)
 
+def parse_llama_print_timings(text: str) -> dict:
+    lines = [n for n in text.splitlines() if n.startswith('llama_print_timings:')]
+    
+    lines = [
+        (
+            '_'.join(line.split('=')[0].split(':')[1].split()),
+            
+            [n.strip() for n in line.split('=')[1].split('(')[0].split('/')] + 
+            [n.strip() for n in line.split('=')[1].split('(')[1].replace(')', '').split(',')]
+            if '/' in line else
+            [n.strip() for n in line.split('=')[1].split('(')[0].split('/')]
+        )
+        for line in lines
+    ]
+
+    lines = dict(lines)
+
+    for k, v in lines.items():
+        lines[k] = {
+            '_'.join(n.split()[1:]): float(n.split()[0])
+            for n in v
+        }
+
+    # print('parse_llama_print_timings:')
+    # print('lines:', lines)
+    return lines
+
 def build_llama_cpp_cmd(device: tuple[int, int, int],
                         prompt: str,
                         model: str,
@@ -157,7 +185,7 @@ async def run_prompt(device: tuple[int, int, int],
     proc_model: str | None = None
     proc: asyncio.subprocess.Process | None = None
 
-    print('? prompt:', repr(prompt))
+    # print('? prompt:', repr(prompt))
 
     cmd: str = build_llama_cpp_cmd(
         device=device,
@@ -194,31 +222,45 @@ async def run_prompt(device: tuple[int, int, int],
             #     devices_procs[index] = (model, proc)
             #     print('devices_procs:', devices_procs)
             # create new proc for model
+            t0: float = time.time()
+
             proc = await asyncio.create_subprocess_shell(
                 cmd,
                 stdout=asyncio.subprocess.PIPE,
                 stderr=asyncio.subprocess.PIPE,
             )
+            
+            stdout: bytes
+            stderr: bytes
 
             if streaming:
+                prev_buf: bytes
                 buf: bytes
                 text: str
 
+                # receive original prompt in stdout
                 # strip original prompt from return
                 while not proc.stdout.at_eof():
                     # stdout
-                    buf = await proc.stdout.read(16)
+                    buf = await proc.stdout.read(1024)
                     stdout += buf
 
                     # skip original prompt
                     if len(stdout) > len(prompt_enc):
                         break
 
+                    await asyncio.sleep(0.1)
+
                 stdout = stdout[1 + len(prompt_enc):]
+                stderr = b''
+                
+                # time-to-load model
+                print('time to load model:', time.time() - t0)
 
                 # return left-overs from stdout as buf
                 buf = stdout
-                text = buf.decode('unicode-escape')
+                prev_buf = b''
+                text = stdout.decode()
 
                 res = {
                     'id': id_,
@@ -231,15 +273,17 @@ async def run_prompt(device: tuple[int, int, int],
 
                 # read rest of tokens
                 while not proc.stdout.at_eof():
-                    buf = await proc.stdout.read(128)
+                    buf = await proc.stdout.read(256)
+                    prev_buf += buf
                     stdout += buf
 
-                    # FIXME: requires better implementation
                     try:
-                        text = buf.decode('unicode-escape')
+                        text = prev_buf.decode()
                     except Exception as e:
                         print('run_prompt buf.decode("unicode-escape") exception:', e)
                         continue
+
+                    prev_buf = b''
                     
                     res = {
                         'id': id_,
@@ -267,6 +311,9 @@ async def run_prompt(device: tuple[int, int, int],
                 stdout, stderr = await proc.communicate()
                 stdout = stdout[1 + len(prompt_enc):]
 
+            if streaming:
+                stderr = await proc.stderr.read()
+
             if stopped:
                 try:
                     proc.kill()
@@ -289,8 +336,14 @@ async def run_prompt(device: tuple[int, int, int],
         finally:
             proc = None
 
-    print('!! stdout', repr(stdout))
-    print('!! stderr', repr(stderr))
+    # print('!! stdout', repr(stdout))
+    # print('!! stderr', repr(stderr))
+
+    print('!! stdout:')
+    print(stdout)
+    
+    print('!! stderr:')
+    print(stderr)
 
     # create eager proc for model
     # proc = await asyncio.create_subprocess_shell(
@@ -347,7 +400,7 @@ async def post_api_1_0_text_completion(request, id_: str):
             print(f'index {index}, platform {pi}, device {di}: available')
             break
         else:
-            await asyncio.sleep(1.0)
+            await asyncio.sleep(0.5)
             continue
 
         if lock:
@@ -384,8 +437,21 @@ async def post_api_1_0_text_completion(request, id_: str):
 
     # post-process
     output = stdout
-    # info = stderr
-    info = None
+    
+    try:
+        llama_timings = parse_llama_print_timings(stderr)
+    except Exception as e:
+        print('parse_llama_print_timings:', e)
+        llama_timings = None
+    
+    if llama_timings:
+        info = {
+            **llama_timings
+        }
+    else:
+        info = {
+            'error': stderr,
+        }
 
     res = {
         'id': id_,
@@ -425,13 +491,13 @@ async def _ws_text_completion_stream(ws, id_: str, data: dict):
                 }
 
                 await ws.send_json(res)
-                await asyncio.sleep(1.0)
+                await asyncio.sleep(0.5)
                 continue
 
             print(f'index {index}, platform {pi}, device {di}: available')
             break
         else:
-            await asyncio.sleep(1.0)
+            await asyncio.sleep(0.5)
             continue
 
         if lock:
@@ -485,8 +551,21 @@ async def _ws_text_completion_stream(ws, id_: str, data: dict):
 
     # post-process
     output = stdout
-    # info = stderr
-    info = None
+
+    try:
+        llama_timings = parse_llama_print_timings(stderr)
+    except Exception as e:
+        print('parse_llama_print_timings:', e)
+        llama_timings = None
+    
+    if llama_timings:
+        info = {
+            **llama_timings
+        }
+    else:
+        info = {
+            'error': stderr,
+        }
 
     res = {
         'id': id_,
